/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package engcalculator.function.prefix.equation;


import engcalculator.function.prefix.PrefixFunctionExpandsLast;
import engcalculator.function.prefix.calculus.*;
import engcalculator.data.Interval;
import engcalculator.data.IntervalList;
import engcalculator.function.CalculusException;
import engcalculator.function.prefix.PrefixFunction;
import engcalculator.function.infix.compare.Equal;
import engcalculator.function.infix.compare.Greater;
import engcalculator.function.infix.compare.Smaller;

/**
 *
 * @author simone
 */

public class FindRoot extends PrefixFunctionExpandsLast {
    private static final Smaller SMALLER = new Smaller();
    private static final Greater GREATER = new Greater();
    private static final Equal EQUAL = new Equal();

    public static final int MAX_NUMBER_OF_ITERATIONS = 1000;
    public static final double EPSILON = 1e-6;

    private final static String[] EXAMPLE = {"'sin', 1_4"};
    private final static String[] RESULT = {"3.142"};

    public FindRoot() {
        super ("findroot", (byte) 2, false, true, "findroot finds the root of a 'thick' function. It needs the equation name and the initial interval containing the root. It will be shrinked to the smallest interval containint the root. It is used bisection method with a maximum of 1000 iterations and epsilo 1e-6. If the function returns multiple variables only the first element will be used to find the root.", EXAMPLE, RESULT);
    }

    private final static Interval ZERO = new Interval (0);

    @Override
    public IntervalList _compute(IntervalList input) throws Exception {
        String fname = input.getFirst().getName();
        if (fname == null) throw new CalculusException ("It is expected a function name as first element");
        PrefixFunction f = PrefixFunction.getFunction(fname);
        if (f == null) throw new CalculusException ("It is not possible to find a function matching the given name");
        if (f.getNumbArgs() != 1) throw new CalculusException ("It is required a function of 1 argument");

        Interval range = input.get(1);
        Interval fmin, fmax, fmid;
        double mid, min, max;
        int findMidMinMaxMixRatio, findMidStep;
        
        fmin = f.compute(new IntervalList(new Interval(range.getLeft()))).getFirst();
        fmax = f.compute(new IntervalList(new Interval(range.getRight()))).getFirst();

        if (! ((SMALLER.compare(fmin, ZERO) && GREATER.compare(fmax, ZERO)) || (GREATER.compare(fmin, ZERO) && SMALLER.compare(fmax, ZERO)))) throw new CalculusException ("The initial interval does not seem to contain a root");

        int maxTrial = MAX_NUMBER_OF_ITERATIONS;

        while (maxTrial > 0 ) {
            mid = range.getVal();
            min = range.getLeft();
            max = range.getRight();
            findMidMinMaxMixRatio = 4;
            findMidStep = 0;
            do {
                --maxTrial;
                fmid = f.compute(new IntervalList(new Interval(mid))).getFirst();
                if (SMALLER.compare(fmid, ZERO)) {
                    if (SMALLER.compare(fmin, ZERO)) {
                        fmin = fmid;
                        min = mid;
                    } else {
                        max = mid;
                        fmax = fmid;
                    }
                    break;
                }
                if (GREATER.compare(fmid, ZERO)) {
                    if (GREATER.compare(fmin, ZERO)) {
                        fmin = fmid;
                        min = mid;
                    } else {
                        max = mid;
                        fmax = fmid;
                    }
                    break;
                }
                if (EQUAL.compare(fmid, ZERO)) {//that works only if the function is thin
                    return new IntervalList (new Interval (mid) );
                }
                switch (findMidStep) {
                    case 2: //restart from left
                        findMidStep = 0;
                        findMidMinMaxMixRatio *= 2;
                    case 0: //left
                        ++findMidStep;
                        mid = ((findMidMinMaxMixRatio-1)*min+max)/findMidMinMaxMixRatio;
                        break;
                    case 1: //right
                         ++findMidStep;
                        mid = ((findMidMinMaxMixRatio-1)*max+min)/findMidMinMaxMixRatio;
                        break;
                }
            } while (maxTrial > 0);
            if (Math.abs(max-range.getRight())+Math.abs(min-range.getLeft()) < EPSILON) break;

            range = new Interval (min, max);
        }
        

        if (maxTrial <= 0) throw new CalculusException("Reached maximum number of iterations");
        return new IntervalList(range);

    }

}
