/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package engcalculator.function.prefix.equation;

import engcalculator.data.Interval;
import engcalculator.data.IntervalList;
import engcalculator.data.MatrixIntervalList;
import engcalculator.function.CalculusException;
import engcalculator.function.prefix.PrefixFunction;
import engcalculator.function.infix.arithmetic.Div;
import engcalculator.function.infix.arithmetic.Mul;
import engcalculator.function.infix.arithmetic.Sub;
import engcalculator.function.prefix.interval.Dual;

/**
 *
 * @author simONE
 */
public class GaussSeidel extends PrefixFunction {
    private final static String[] EXAMPLE = {"(10, 2, 3, 40)#2,(5,6)#1,(1,1)#1"};
    private final static String[] RESULT = {"(0.47716, 0.11421)#1"};
    private final static Mul MUL = new Mul();
    private final static Div DIV = new Div ();
    private final static Sub SUB = new Sub();
    private final static Dual DUAL = new Dual();
    private final static double PRECISION = 1e-4;

    public GaussSeidel() {
        super("gaussseidel", (byte) -1, false, true, "Diagonalize a diagonal dominant matrixusing Gauss seidel algorithm. The matrix must be diagona dominat to ensure the methode convergence", EXAMPLE, RESULT);
    }

    @Override
    public IntervalList _compute(IntervalList input) throws Exception {
        if (! input.isProperMatrix() && input.getRows() != input.getColumns()-2) throw new CalculusException ("The input matrix for a system should have n rows and n+1 columns");
        final int rows = input.getRows();
        int iterations = 1000;

        Interval acc, prex, newx;
        double error=10;

        while(iterations-- > 0 && error > PRECISION) {
            error = 0;

            for(int r = 0; r < rows; r++) {
                prex = new Interval(input.get(r, rows+1)); //store previous x component for stop condition
                acc = input.get(r, rows);
                for(int c = 0; c < rows; c++) {
                    if( r != c ) {
                        acc = SUB._computeIntervals(acc, DUAL._computeIntervals(MUL._computeIntervals(input.get(r, c), input.get(c, rows+1))));
                    }
                }
                newx = DIV._computeIntervals(acc, DUAL._computeIntervals(input.get(r, r)));
                input.set(r, rows+1, newx);
                error += Math.abs(newx.getLeft()-prex.getLeft())+Math.abs(newx.getRight()-prex.getRight());
            }
        }

        MatrixIntervalList result = new MatrixIntervalList(1);
        for (int r=0; r<rows; ++r)
            result.append(input.get(r, rows+1));
        
        return result;
    }

}
